#ifndef ARRAY2D_H
#define ARRAY2D_H

#include "Row.h"

template <typename T>
class Array2D
{ 
public:
	// 4 default c'tors
	Array2D( int row, int column );
	~Array2D();
	Array2D( const Array2D< T > & copy );
	Array2D< T > & operator = ( const Array2D< T > & copy );

	// selection methods
	T & Select( int row, int column );
	Row< T > operator [] ( int row );

    //mutators
	void SetRows( int );
	void SetColumns( int );

	int GetRows() const;
	int GetColumns() const;
private:
	T** m_array;
	int m_rows;
	int m_columns;
};

template <typename T>
Array2D< T >::Array2D( int row, int column )
{
	if( row <= 0 )
		throw OutOfBounds( "ERROR: cannot have negative or 0 rows" );

	if( column <= 0 )
		throw OutOfBounds( "ERROR: cannot have negative or 0 columns" );

	m_rows = row;
	m_columns = column;
	m_array = new T*[ row ];

	for( int i = 0; i < row; i++ )
		m_array[i] = new T[column];
}

template <typename T>
Array2D< T >::~Array2D()
{
	for( int i = 0; i < m_rows; i++ )
		delete [] m_array[i];

	delete [] m_array;
}

template <typename T>
Array2D< T >::Array2D( const Array2D< T > & copy ):m_array(0)
{
	*this = copy;
}

template <typename T>
Array2D< T > & Array2D< T >::operator = ( const Array2D< T > & copy )
{
	if( this != &copy )
	{
		if( copy.m_array )
		{
			if( m_array )
			{
				for( int i = 0; i < m_rows; i++ )
					delete [] m_array[i];

				delete [] m_array;
			}

			m_rows = copy.m_rows;
			m_columns = copy.m_columns;

			m_array = new T*[ m_rows ];

			for( int i = 0; i < m_rows; i++ )
			{
				m_array[i] = new T[ m_columns ];

				for( int j = 0; j < m_columns; j++ )
					m_array[i][j] = copy.m_array[i][j];
			}
		}
		else
		{
			m_rows = copy.m_rows;
			m_columns = copy.m_columns;
		}
	}
	
	return *this;
}

template <typename T>
T & Array2D< T >::Select( int row, int column )
{
	if( row < 0 || row >= m_rows )
		throw OutOfBounds( "ERROR: Row index out of bounds." );

	if( column < 0 || column >= m_columns )
		throw OutOfBounds( "ERROR: Column index out of bounds." );

	return m_array[row][column];
}

template <typename T>
Row< T > Array2D< T >::operator []( int row )
{
	if( row < 0 || row >= m_rows )
		throw OutOfBounds( "ERROR: Row index out of bounds." );

	return Row< T >( *this, row );
}

template <typename T>
void Array2D< T >::SetRows( int row )
{
	if( row <= 0 )
		throw OutOfBounds( "ERROR: cannot have negative or 0 rows" );

	if( m_rows == row ) // if equal, no change, so return
		return;

	// make temp array for new size
	T** temp = new T*[ row ];

	if( m_rows < row ) // if adding more rows
	{
		// copy all the rows over to temp
		for( int i = 0; i < m_rows; i++ )
			temp[i] = m_array[i];

		// make the additional rows in temp
		for( int i = m_rows; i < row; i++ )
			temp[i] = new T[ m_columns ];
	}
	else // if subtracting rows
	{
		// copy all the rows over to temp up to the number of new rows
		for( int i = 0; i < row; i++ )
			temp[i] = m_array[i];

		// delete excess rows in origonal array m_array
		for( int i = row; i < m_rows; i++ )
			delete [] m_array[i];
	}

	delete [] m_array;
	m_array = temp;
	m_rows = row;
}

template <typename T>
void Array2D< T >::SetColumns( int col )
{
	if( col <= 0 )
		throw OutOfBounds( "ERROR: cannot have negative or 0 columns" );

	if( m_columns == col ) // if equal, no change, so return
		return;

	// make temp array for new size

	T** temp = new T*[ m_rows ];

	for( int i = 0; i < m_rows; i++ )
		temp[i] = new T[ col ];

	if( m_columns > col )
		m_columns = col;

	for( int i = 0; i < m_rows; i++ )
	{
		// copy all data elements over to temp
		for( int j = 0; j < m_columns; j++ )
			temp[i][j] = m_array[i][j];

		// delete old row
		delete [] m_array[i];
		// assign temp row to our row
		m_array[i] = temp[i];
	}

	// delete temp array of pointers; they are no longer needed
	delete [] temp;
	m_columns = col;
}

template <typename T>
int Array2D< T >::GetRows() const
{
	return m_rows;
}

template <typename T>
int Array2D< T >::GetColumns() const
{
	return m_columns;
}

#endif