/***********************************************************
* Author:				Jonathan Lavering
* Date Created:			4/3/2004
* Last Modification Date:	4/8/2004
* Lab Number:			CST 211 Lab 1
* Filename:				main.cpp
*
* Overview:
*	This program simulates the game of Minesweeper, such as
*	the one in Windows.  A random grid of covered squares
*	hide a minefield, where the user must uncover all the
*	squares except the ones the mines are under.  Number
*	squares identify how many mines surround that square.
*
* Input:
*	Input comes from the user of the program, through the
*	standard console.  The user will be asked to specify
*	the difficulty he/she wishes to play and then must
*	repeatedly enter the grid coordinates that he/she wishes
*	to uncover or mark/unmark.  The program handles most
*	incorrect inputs.
*
* Output:
*	The program continually updates the screen after every
*	user action.  The game will end when either the user wins
*	or loses.  The user can win if he/she uncovers all the
*	squares on the grid except for those that hide mines.
*	The user will lose if he/she uncovers a square that hides
*	a mine.  The program will then ask if the user wishes to
*	continue, and if so, a new game is started, otherwise,
*	the program exits.
************************************************************/

// includes
#include <iostream>
#include <cstdlib>
#include <windows.h>
#include <ctime>
#include <iomanip>
#include "Array2D.h"
#include "OutOfBounds.h"

using namespace std;

// function prototypes

// This function resets the user grid to all covered squares
Array2D< char > ResetUserGrid( char level );

// This function generates a new, random base grid of empty spaces, mines, and number squares
Array2D< char > GenerateRandGrid( char level );

// This function displays the current user gird
void Display( Array2D< char > userGrid );

// This function gets the difficulty level from the user
char GetLevel( void );

// This function increments the all the number squares surrounding a mine by 1
void IncrementNumSqr( int row, int col, Array2D< char > & baseGrid );

// This function gets a new grid coord to uncover or mark/unmark from the user.
// Returns true if marking a the square, false if not
bool GetInput( int & row, int & col, Array2D< char > userGrid );

// Uncovers current square; returns true if uncovered bomb, false if not
bool UncoverSquare( int row, int col, Array2D< char > & userGrid, Array2D< char > baseGrid );

// Markes or unmarks current covered square
void MarkSquare( int row, int col, Array2D< char > & userGrid );

// Checks if the user won or not; returns true if so, false if not
bool CheckIfWon( char level, Array2D< char > userGrid );

// Function clears the console screen
void clrscr();

void main()
{
	// program variables and flags
	Array2D< char >userGrid( 10, 10 ); // grid user sees
	Array2D< char >baseGrid( 10, 10 ); // actual grid that shows all mines
	char level = 'b'; // difficulty level { beginner, intermediate, and expert }
	char choice = 'y'; // flag for the user to choose to play another game
	int row = 0; // storage for current row position
	int col = 0; // storage for current column position
	bool marking = false; // flag to show if the user wishes to mark a square
	bool endGame = false; // flag to show if the game has ended
	bool wonGame = false; // flag to show if the user won the game

	srand( time( 0 ) ); // randomize random function generator

	cout << "Welcome to Minesweeper!" << endl;

	while( choice == 'y' ) // while loop for the user to keep playing if he/she wishes
	{
		clrscr(); // clear the screen
		level = GetLevel(); // get difficulty level by user
		userGrid = ResetUserGrid( level ); // reset the user grid to being fully covered
		// generate a random base grid with mines, empty spaces, and numbers
		baseGrid = GenerateRandGrid( level );

		// base game loop; loop while the game hasn't ended and while game hasn't been won
		while( ( endGame == false ) && ( wonGame == false ) )
		{
			clrscr(); // clear the screen
			Display( userGrid ); // display the current grid as seen by the user
			marking = GetInput( row, col, userGrid ); // get the current input; determine if marking or not

			if( marking ) // if user is marking a square
			{
				MarkSquare( row, col, userGrid ); // mark/unmark the square
				marking = false; // reset the marking flag
			}
			else // else, if uncovering a square
			{
				// uncover the square, and all surrounding it up to a number square.
				// return true of the user uncovered a mine square, false if not
				endGame = UncoverSquare( row, col, userGrid, baseGrid );

				if( !endGame ) // if the user didn't uncover a mine
					wonGame = CheckIfWon( level, userGrid ); // check if he/she won
			}
		}

		clrscr();
		Display( userGrid );

		if( wonGame )
		{
			cout << "Congratulations! You've won!" << endl;
			wonGame = false;
		}
		else
		{
			cout << "Sorry, you've lost." << endl;
			endGame = false;
		}

		cout << "Play again? (y/n): ";
		cin >> choice;
	}
}

/**************************************************************
*   Entry:  none
*
*    Exit:  fills a char grid with '?'.
*
* Purpose:  Resets the user grid to all covered squares '?'.
*
***************************************************************/
Array2D< char > ResetUserGrid( char level )
{
	Array2D< char >grid( 10, 10 ); // temp grid

	switch( level ) // determine which difficulty was chosen
	{
	case 'b': // beginner
		grid.SetRows( 10 );
		grid.SetColumns( 10 );
		break;
	case 'i': // intermediate
		grid.SetRows( 16 );
		grid.SetColumns( 16 );
		break;
	case 'e': // expert
		grid.SetRows( 16 );
		grid.SetColumns( 30 );
	};

	// fill grid with covered squares '?'
	for( int i = 0; i < grid.GetRows(); i++ )
		for( int j = 0; j < grid.GetColumns(); j++ )
			grid[i][j] = '?';

	return grid;
}

/**************************************************************
*   Entry:  none
*
*    Exit:  First, fills a temp char grid with '_', then
*			randomly places mines 'X' in it, and then produces
*			the correct number squares surrounding each mine.
*
* Purpose:  To generate a random base grid for a user to to
*			work off of from a covered version of it.
*
***************************************************************/
Array2D< char > GenerateRandGrid( char level )
{
	Array2D< char >grid( 10, 10 ); // temp grid
	int mineCount; // the number of mines needed to place on the grid
	int mineElement; // element number randomly chosen to place the mine
	int mineRow; // the row that the randomly chosen element is on
	int mineCol; // the column that the randomly chosen element is on
	int gridSize; // size of the grid

	switch( level ) // determine which difficulty was chosen
	{
	case 'b': // beginner
		grid.SetColumns( 10 );
		grid.SetRows( 10 );
		mineCount = 10;
		gridSize = 100; // 10 X 10
		break;
	case 'i': // intermediate
		grid.SetColumns( 16 );
		grid.SetRows( 16 );
		mineCount = 40;
		gridSize = 256; // 16 X 16
		break;
	case 'e': // expert
		grid.SetColumns( 30 );
		grid.SetRows( 16 );
		mineCount = 100;
		gridSize = 480; // 16 X 30
	};

	// fill the gird with empty squares '_'
	for( int i = 0; i < grid.GetRows(); i++ )
		for( int j = 0; j < grid.GetColumns(); j++ )
			grid[i][j] = '_';

	// randomly place the mines in the grid 'X'
	while( mineCount != 0 )
	{
		mineElement = rand() % gridSize;
		mineCol = mineElement % grid.GetColumns();
		mineRow = ( mineElement - mineCol ) / grid.GetColumns() ;
		
		if( grid[ mineRow ][ mineCol ] != 'X' )
		{
			grid[ mineRow ][ mineCol ] = 'X';
			mineCount--;
		}
	}

	// search through grid and generate the number squares
	for( int i = 0; i < grid.GetRows(); i++ )
	{
		for( int j = 0; j < grid.GetColumns(); j++ )
		{
			// if current square is a bomb,
			// increase the numbers around it
			if( grid[i][j] == 'X' )
			{
				// square up-left
				IncrementNumSqr( i - 1, j - 1, grid );
				// square up-mid
				IncrementNumSqr( i - 1, j, grid );
				// square up-right
				IncrementNumSqr( i - 1, j + 1, grid );
				// square mid-left
				IncrementNumSqr( i, j - 1, grid );
				// square mid-right
				IncrementNumSqr( i, j + 1, grid );
				// square down-left
				IncrementNumSqr( i + 1, j - 1, grid );
				// square down-mid
				IncrementNumSqr( i + 1, j, grid );
				// square down-right
				IncrementNumSqr( i + 1, j + 1, grid );
			}
		}
	}

	return grid;
}

/**************************************************************
*   Entry:  none
*
*    Exit:  Prints row and column numbers on top and left edge
*			of grid, and then prints out the array grid itself.
*
* Purpose:  To display the current user grid.
*
***************************************************************/
void Display( Array2D< char > userGrid )
{
	// show column numbers
	cout << "   ";

	for( int i = 1; i <= userGrid.GetColumns(); i++ )
		cout << left << setw(2) << i;

	cout << endl << endl;

	for( int i = 0; i < userGrid.GetRows(); i++ )
	{
		// show row numbers
		cout << left << setw(3) << i + 1;

		// print the current grid position
		for( int j = 0; j < userGrid.GetColumns(); j++ )
			cout << userGrid[i][j] << " ";

		cout << endl;
	}

	cout << endl;
}

/**************************************************************
*   Entry:  Called only when starting a new game.
*
*    Exit:  Asks user for a difficulty of beginner, intermediate
*			or expert (b, i, e).  Will repeatedly ask until user
*			inputs a valid input of b, i, or e.
*
* Purpose:  To get the difficulty setting of the new game.
*
***************************************************************/
char GetLevel( void )
{
	char level; // holds difficulty level as a char: b, i, e

	cout << "Enter difficulty level:" << endl
		 << "	(b)eginner" << endl
		 << "	(i)ntermediate" << endl
		 << "	(e)xpert" << endl;
	cin >> level;

	// re-ask the user for a correct difficulty setting if he/she didn't input b, i, or e
	while( level != 'b' && level != 'i' && level != 'e' )
	{
		cout << "Please enter either characters 'b', 'i', or 'e' for beginner," << endl
			 << " intermediate, or expert levels: ";
		cin >> level;
	}
	
	return level;
}

/**************************************************************
*   Entry:  Called only in GenerateRandGrid().
*
*    Exit:  Checks if the current square is within array bounds
*			and if not a mine.  If this condition is true, the
*			square is incremented to a number if currently an
*			emtpy one, or just increments the current number.
*
* Purpose:  Used to increment the number value of the squares
*			surrounding a mine to eventually produce a number
*			that shows how many mines are around that number
*			square.
*
***************************************************************/
void IncrementNumSqr( int row, int col, Array2D< char > & baseGrid )
{
	bool legal_square = false; // flag that tells if the current square is legal

	// if square is within array bounds
	if( ( row >= 0 && row < baseGrid.GetRows() ) && 
		( col >= 0 && col < baseGrid.GetColumns() ) &&
		( baseGrid[row][col] != 'X' ) )
		legal_square = true; // it's a legal square
	else
		legal_square = false; // otherwise, its not

	if( legal_square ) // if it is a legal square
	{
		// increment the square number
		if( baseGrid[row][col] == '_' )
			baseGrid[row][col] = '1';
		else
			baseGrid[row][col]++;
	}
}

/**************************************************************
*   Entry:  Called only at the beginning of a new game.
*
*    Exit:  User must enter a valid input coord for
*			uncovering a square or marking/unmarking a square.
*
* Purpose:  Gets the coord that the user wishes to uncover or
*			mark/unmark.
*
***************************************************************/
bool GetInput( int & row, int & col, Array2D< char > grid )
{
	bool legalinput = false;
	bool marking = false;

	while( !legalinput )
	{
		cout << "Covered = ? Blank = _ Marked = * Bomb = X" << endl;
		cout << "Enter coord (row column); use (-row -column) to mark: ";
		cin >> row >> col;

		// legal to input complement coord for marking
		if( row < 0 && col < 0 ) // ( -row -col )
		{
			row = row * -1;
			col = col * -1;

			// must be within array bounds
			if( ( row - 1 >= grid.GetRows() ) || ( col - 1 >= grid.GetColumns() ) )
				legalinput = false;
			else
			{
				if( ( grid[row - 1][col - 1] != '?' ) &&
					( grid[row - 1][col - 1] != '*' ) )
					legalinput = false;
				else
				{
					legalinput = true;
					marking = true;
				}
			}
		}
		else if( row > 0 && col > 0 ) // ( row col )
		{
			// must be within array bounds
			if( ( ( row - 1 ) >= grid.GetRows() ) || ( ( col - 1 ) >= grid.GetColumns() ) )
				legalinput = false;
			else
			{
				if( grid[row - 1][col - 1] != '?' ) // can only click on '?' squares
					legalinput = false;
				else
					legalinput = true;
			}
		}
		else
			legalinput = false;

		if( !legalinput )
		{
			clrscr();
			Display( grid );
		}
	}

	// decrement row and col variables for use in array indexing
	row = row - 1;
	col = col - 1;
	return marking;
}

/**************************************************************
*   Entry:  Called during game loop.  Must only be called when
*			not marking/unmarking a square.
*
*    Exit:  Exits if the square is a mine - returns true for
*			endgame flag.  Exits if the square is already
*			uncovered but returns false to show that its not
*			endgame.  Exits if the previous two conditions fail
*			by uncovering all the squares in every direction up
*			to a number square.
*
* Purpose:  To uncover the current square.  The game will end if
*			the square is a mine, otherwise, it will try to
*			uncover as much space as possible up to the
*			surrounding number squares.
*
***************************************************************/
bool UncoverSquare( int row, int col, Array2D< char > & userGrid, Array2D< char > baseGrid )
{
	if( baseGrid[row][col] == 'X' )
	{
		userGrid = baseGrid; // reveal grid
		return true; // game ends
	}

	if( userGrid[row][col] != '?' ) // if not a covered square
		return false; // cannot uncover already uncovered square, but game still continues
	else // else if a covered square
	{
		userGrid[row][col] = baseGrid[row][col]; // uncover the square

		if( baseGrid[row][col] != '_' ) // if not an empty square
			return false; // stop at number square and return
		else
		{
			// try uncovering square up-left
			try { UncoverSquare( row - 1, col - 1, userGrid, baseGrid ); } catch( OutOfBounds excep ){ }

			// try uncovering square up-mid
			try{ UncoverSquare( row - 1, col, userGrid, baseGrid ); } catch( OutOfBounds excep ){ }

			// try uncovering square up-right
			try{ UncoverSquare( row - 1, col + 1, userGrid, baseGrid ); } catch( OutOfBounds excep ){ }

			// try uncovering square left-mid
			try{ UncoverSquare( row, col - 1, userGrid, baseGrid ); } catch( OutOfBounds excep ){ }

			// try uncovering square right-mid
			try{ UncoverSquare( row, col + 1, userGrid, baseGrid ); } catch( OutOfBounds excep ){ }

			// try uncovering square down-left
			try{ UncoverSquare( row + 1, col - 1, userGrid, baseGrid ); } catch( OutOfBounds excep ){ }

			// try uncovering square down-mid
			try{ UncoverSquare( row + 1, col, userGrid, baseGrid ); } catch( OutOfBounds excep ){ }

			// try uncovering square down-right
			try{ UncoverSquare( row + 1, col + 1, userGrid, baseGrid ); } catch( OutOfBounds excep ){ }

			return false;
		}
	}
}

/**************************************************************
*   Entry:  Called only if the user wishes to mark the current
*			square.
*
*    Exit:  If the square is covered, mark, otherwise, if it's
*			marked, unmark it.
*
* Purpose:  To mark or unmark the current square.
*
***************************************************************/
void MarkSquare( int row, int col, Array2D< char > & userGrid )
{
	// if square is already marked, unmark it
	if( userGrid[row][col] == '*' )
		userGrid[row][col] = '?';
	else // else, mark it
		userGrid[row][col] = '*';
}

/**************************************************************
*   Entry:  Called only if uncovering a square, not marking it.
*			Called right after UncoverSquare() is called.
*
*    Exit:  Counts up the total number of covered and marked
*			squares.  if the sum of the total number of marked
*			and covered squares is equal to the number of mines
*			on the grid, the user has won.
*
* Purpose:  To determine if the user has won the game or not.
*
***************************************************************/
bool CheckIfWon( char level, Array2D< char > userGrid )
{
	int coveredSquares = 0; // stores the number of covered squares on the gird
	int markedSquares = 0; // stores the number of marked squares on the grid
	int numMines = 0; // stores the number of mines on the grid

	switch( level ) // determine which difficulty is active
	{
	case 'b': // beginner
		numMines = 10;
		break;
	case 'i': // intermediate
		numMines = 40;
		break;
	case 'e': // expert
		numMines = 100;
	}

	// count how many covered and marked squares there are
	for( int i = 0; i < userGrid.GetRows(); i++ )
	{
		for ( int j = 0; j < userGrid.GetRows(); j++ )
		{
			if( userGrid[i][j] == '?' )
				coveredSquares++;
			else if( userGrid[i][j] == '*' )
				markedSquares++;
		}
	}

	// if the sum of the covered and marked squares 
	// is equal to the number of mines, the user won
	if( coveredSquares + markedSquares == numMines )
		return true;
	else
		return false;
}

/**************************************************************
*   Entry:  none
*
*    Exit:  Fills the console with ' ' empty spaces.
*
* Purpose:  To clear the console.
*
***************************************************************/
void clrscr()
{
	// by tequilla---
	// found on    http://cpp.enisoc.com/articles/clrscr/

	HANDLE hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	COORD coord = {0, 0};
	DWORD count;

	CONSOLE_SCREEN_BUFFER_INFO csbi;
	GetConsoleScreenBufferInfo(hStdOut, &csbi);

	FillConsoleOutputCharacter(hStdOut, ' ', csbi.dwSize.X * csbi.dwSize.Y, coord, &count);

	SetConsoleCursorPosition(hStdOut, coord);
}